import arcgis as _arcgis 
from ._arcgis_model import ArcGISModel
import random
try:
    import pandas
    import tempfile
    import numpy as np
    import json
    import os
    import io
    import shutil
    import warnings
    from pathlib import Path
    from functools import partial
    from ._unet_utils import is_no_color
    from ._codetemplate import feature_classifier_prf
    import torch
    from torchvision import models
    from fastai.metrics import accuracy
    from fastai.vision.image import open_image
    from fastai.vision.data import ImageDataBunch, ImageList
    from fastai.vision import imagenet_stats, normalize
    from fastai.vision.learner import cnn_learner, ClassificationInterpretation
    from ._arcgis_model import _set_multigpu_callback
    from fastai.vision.transform import crop, rotate, dihedral_affine, brightness, contrast, skew, rand_zoom, get_transforms
    import torch.nn.functional as functional
    from .._data import _check_esri_files
    import glob
    import time
    import xml.etree.ElementTree as ElementTree
    import PIL.Image
    import PIL.ExifTags
    from torch.nn import Module as NnModule
    HAS_FASTAI = True
except Exception as e:
    class NnModule():
        pass
    HAS_FASTAI = False


def _mobilenet_split(m:NnModule): return m[0][0][0], m[1]


def _prediction_function(predictions):
    classes = {}
    max_prediction_value = 0
    max_prediction_class = None
    for prediction in predictions:
        if not classes.get(prediction[0]):
            classes[prediction[0]] = prediction[1]
        else:
            classes[prediction[0]] = classes[prediction[0]] + prediction[1]
        if max_prediction_value < classes[prediction[0]]:
            max_prediction_value = classes[prediction[0]]
            max_prediction_class = prediction[0]

    return max_prediction_class, max_prediction_value


class FeatureClassifier(ArcGISModel):
    """
    Creates an image classifier to classify the area occupied by a
    geographical feature based on the imagery it overlaps with.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional torchvision model. Backbone CNN model to be used for
                            creating the base of the ``FeatureClassifier``, which
                            is ``resnet34`` by default.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :returns: `FeatureClassifier` Object
    """

    def __init__(self, data, backbone=None, pretrained_path=None):
        
        super().__init__(data, backbone)

        backbone_cut = None
        backbone_split = None
        if self._backbone == models.mobilenet_v2:
            backbone_cut = -1
            backbone_split = _mobilenet_split

        if not self._check_backbone_support(self._backbone):
            raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

        self._code = feature_classifier_prf
        self.learn = cnn_learner(data, self._backbone, metrics=accuracy, cut=backbone_cut, split_on=backbone_split)

        self.learn.model = self.learn.model.to(self._device)

        _set_multigpu_callback(self)
        if pretrained_path is not None:
            self.load(pretrained_path)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        return [*self._resnet_family, models.mobilenet_v2.__name__]

    def show_results(self, rows=5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        if rows > self._data.batch_size:
            rows = self._data.batch_size
        self.learn.show_results(rows=rows, **kwargs)

    def predict(self, img_path):
        img = open_image(img_path)
        return self.learn.predict(img)

    def _predict_batch(self, imagetensor_batch):
        predictions = self.learn.model.eval()(imagetensor_batch.to(self._device)).detach().cpu()
        predictions_conf,  predicted_classes = torch.max(predictions, dim=-1)
        predicted_classes = predicted_classes.tolist()
        predictions_conf = (predictions_conf*100).tolist()
        return predicted_classes, predictions_conf

    def _save_confusion_matrix(self, path):
        from matplotlib import pyplot as plt
        import fastai
        import fastprogress
        fastprogress.fastprogress.NO_BAR = True
        fastai.basic_train.master_bar, fastai.basic_train.progress_bar = fastprogress.force_console_behavior()
        self.plot_confusion_matrix()
        fastai.basic_train.master_bar, fastai.basic_train.progress_bar = fastprogress.master_bar, fastprogress.progress_bar
        plt.savefig(os.path.join(path, 'confusion_matrix.png'))
        plt.close()

    @property
    def _model_metrics(self):
        return {}
    
    def _create_emd(self, path):
        super()._create_emd(path)
        self._emd_template["Framework"] = "PyTorch"
        self._emd_template["ModelConfiguration"] = "FeatureClassifier"
        self._emd_template["ModelType"] = "ObjectClassification"
        self._emd_template["ExtractBands"] = [0, 1, 2]
        self._emd_template['CropSizeFixed'] = 1  # hardcoded
        self._emd_template['BlackenAroundFeature'] = 0 #hardcoded
        self._emd_template['ImageSpaceUsed'] = "MAP_SPACE"
        self._emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes):
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())

        json.dump(self._emd_template, open(path.with_suffix('.emd'), 'w'), indent=4)

        return path.stem

    def _create_tf_emd(self, saved_path, onnx_path):
        import random
        super()._create_tf_emd(saved_path, onnx_path)

        self._emd_template["Framework"] = "arcgis.learn.models._inferencing"
        self._emd_template["InferenceFunction"] = "ArcGISObjectDetector.py"
        self._emd_template["ModelConfiguration"] = "_classifier"
        self._emd_template["ExtractBands"] = [0, 1, 2]
        self._emd_template['Classes'] = []

        class_data = {}
        for i, class_name in enumerate(self._data.classes):
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            self._emd_template['Classes'].append(class_data.copy())

        json.dump(self._emd_template, open(saved_path.with_suffix('.emd'), 'w'), indent=4)

        return saved_path.stem

    @classmethod
    def from_model(cls, emd_path, data=None):
        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd['ModelParameters']
        chip_size = emd["ImageWidth"]

        try:
            class_mapping = {i['Value']: i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value']: i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue']: i['ClassName'] for i in emd['Classes']}
            color_mapping = {i['ClassValue']: i['Color'] for i in emd['Classes']}

        if data is None:
            ranges = (0, 1)
            train_tfms = [rotate(degrees=30, p=0.5),
                          crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges),
                          dihedral_affine(), brightness(change=(0.4, 0.6)), contrast(scale=(0.75, 1.5)),
                          # rand_zoom(scale=(0.75, 1.5))
                          ]
            val_tfms = [crop(size=chip_size, p=1.0, row_pct=0.5, col_pct=0.5)]
            transforms = (train_tfms, val_tfms)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)

                tempdata = ImageDataBunch.single_from_classes(
                    tempfile.TemporaryDirectory().name, sorted(list(class_mapping.values())),
                    ds_tfms=transforms, size=chip_size).normalize(imagenet_stats)
                tempdata.chip_size = chip_size
                tempdata.class_mapping = class_mapping
                tempdata.classes = list(class_mapping.keys())
                data = tempdata

        resize_to = emd.get('resize_to')
        data.resize_to = resize_to

        return cls(data, **model_params, pretrained_path=str(model_file))

    def plot_confusion_matrix(self):
        """
        Plots a confusion matrix of the model predictions to evaluate accuracy
        """
        interp = ClassificationInterpretation.from_learner(self.learn)
        interp.plot_confusion_matrix()

    def plot_hard_examples(self, num_examples):
        """
        Plots the hard examples with their heatmaps.
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        num_examples            Number of hard examples to plot
                                ``prepare_data`` function.
        =====================   ===========================================
        """
        interp = ClassificationInterpretation.from_learner(self.learn)
        interp.plot_top_losses(num_examples, figsize=(15,15), heatmap=True)

    @staticmethod
    def _convert_to_degrees(value, reference):
        d0 = value[0][0]
        d1 = value[0][1]
        d = float(d0) / float(d1)

        m0 = value[1][0]
        m1 = value[1][1]
        m = float(m0) / float(m1)

        s0 = value[2][0]
        s1 = value[2][1]
        s = float(s0) / float(s1)

        degrees = d + (m / 60.0) + (s / 3600.0)

        if reference == "S" or reference == "W":
            degrees = 0 - degrees
        
        return degrees

    def predict_folder_and_create_layer(self, folder, feature_layer_name, gis=None, prediction_field='predict', confidence_field='confidence'):
        """
        Predicts on images present in the given folder and creates a feature layer.
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        folder                  Required String. Folder to inference on.
        ---------------------   -------------------------------------------
        feature_layer_name      Required String. The name of the feature layer used to publish.   
        ---------------------   -------------------------------------------
        gis                     Optional GIS Object, the GIS on which this tool runs. If not specified,
                                the active GIS is used.      
        ---------------------   -------------------------------------------
        prediction_field        Optional String. The field name to use to add predictions.             
        ---------------------   -------------------------------------------
        confidence_field        Optional String. The field name to use to add confidence.                                                                               
        =====================   ===========================================

        :returns: `FeatureCollection` Object                              
        """        
        return self._create_feature_layer(
            self._extract_images_geo_data(folder),
            _arcgis.env.active_gis if gis is None else gis,
            feature_layer_name,
            prediction_field,
            confidence_field
        )

    def _extract_images_geo_data(self, folder):
        ALLOWED_FILE_FORMATS = ['tif', 'jpg', 'png']

        files = []

        for ext in ALLOWED_FILE_FORMATS:
            files.extend(glob.glob(os.path.join(folder, '*.' + ext)))
        
        images_data = []

        for file in files:
            img = PIL.Image.open(file)
            exif = {
                PIL.ExifTags.TAGS[k]: v
                for k, v in img._getexif().items()
                if k in PIL.ExifTags.TAGS
            }

            images_data.append(
                {
                    'image_path': file,
                    'y': FeatureClassifier._convert_to_degrees(exif['GPSInfo'][2], exif['GPSInfo'][1]),
                    'x': FeatureClassifier._convert_to_degrees(exif['GPSInfo'][4], exif['GPSInfo'][3])
                }
            )
        
        return images_data

    def _create_feature_layer(self, images_data, gis_user, feature_layer_name, prediction_field, confidence_field):
        data = []
        images = {}
        for image_data in images_data:
            prediction = self.predict(image_data['image_path'])
            images[os.path.basename(image_data['image_path'])] = image_data['image_path']
            data.append(
                [   
                    os.path.basename(image_data['image_path']),
                    prediction[0].obj,
                    prediction[2].data.max().tolist(),
                    image_data['x'],
                    image_data['y']
                ]
            )
        
        dataframe = pandas.DataFrame(data, columns=['Image_Name', prediction_field, confidence_field, 'X', 'Y'])
        spatial_dataframe = dataframe.spatial.from_xy(df=dataframe, sr=4326, x_column='X', y_column='Y')

        feature_collection = gis_user.content.import_data(spatial_dataframe, title=feature_layer_name)

        feature_layer = feature_collection.layers[0]
        feature_layer.manager.add_to_definition({"hasAttachments":True})

        df = feature_layer.query(as_df=True)
        object_field = feature_layer.properties['objectIdField']

        for image_name, image_path in images.items():
            object_id = df[object_field].where(df['Image_Name'] == image_name).values[0]  #assuming image_name is unique
            if np.isnan(object_id):
                continue #skipping those values which are not present.
            feature_layer.attachments.add(
                object_id,
                image_path
            )

        return feature_collection

    @staticmethod
    def _update_predictions_layer(feature_layer, features_to_update, output_label_field, confidence_field=None):
        field_template = {
            "name": output_label_field,
            "type": "esriFieldTypeString",
            "alias": output_label_field,
            "sqlType": "sqlTypeOther",
            "length": 256,
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": ''
        }

        confidence_field_template = {
            "name": confidence_field,
            "type": "esriFieldTypeString",
            "alias": confidence_field,
            "sqlType": "sqlTypeOther",
            "length": 256,
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": ''
        }

        feature_layer.manager.add_to_definition({'fields': [field_template]})

        if confidence_field:
            feature_layer.manager.add_to_definition({'fields': [confidence_field_template]})

        try:
            start = 0
            stop = 100
            count = 100

            features_updated = features_to_update[start:stop]
            response = feature_layer.edit_features(updates=features_updated)

            for resp in response.get('updateResults', []):
                if resp.get('success', False):
                    continue
                warnings.warn(f"Something went wrong for data {resp}")

            time.sleep(2)
            while count == len(features_updated):
                start = stop
                stop = stop + 100
                features_updated = features_to_update[start:stop]
                response = feature_layer.edit_features(updates=features_updated)
                for resp in response.get('updateResults', []):
                    if resp.get('success', False):
                        continue
                    warnings.warn(f"Something went wrong for data {resp}")
                time.sleep(2)
        except Exception:
            feature_layer.manager.delete_from_definition({'fields': [field_template]})
            if confidence_field:
                feature_layer.manager.delete_from_definition({'fields': [confidence_field_template]})

            return False

        return True

    def _classify_attachments(
            self,
            feature_layer,
            data_folder,
            feature_attachments_mapping,
            input_label_field,
            output_label_field,
            confidence_field=None,
            predict_function=_prediction_function
    ):

        features = feature_layer.query().features
        features_to_update = []

        for feature in features:
            feature_attachments = (feature_attachments_mapping.get(str(feature.attributes[input_label_field])) or \
                                   feature_attachments_mapping.get(int(feature.attributes[input_label_field])))
            if not feature_attachments:
                continue

            predictions = []
            for attachment in feature_attachments:
                prediction = self.predict(os.path.join(data_folder, attachment))
                predictions.append((prediction[0].obj, prediction[2].data.max().tolist()))

            final_prediction = predict_function(predictions)

            feature.attributes[output_label_field] = final_prediction[0]
            if confidence_field:
                feature.attributes[confidence_field] = final_prediction[1]

            features_to_update.append(feature)

        return features_to_update

    def _classify_labeled_tiles(
            self,
            feature_layer,
            labeled_tiles_directory,
            input_label_field,
            output_label_field,
            confidence_field=None
    ):
        ALLOWED_FILE_FORMATS = ['tif', 'jpg', 'png']
        IMAGES_FOLDER = 'images/'
        LABELS_FOLDER = 'labels/'

        files = []

        for ext in ALLOWED_FILE_FORMATS:
            files.extend(glob.glob(os.path.join(labeled_tiles_directory, IMAGES_FOLDER + '*.' + ext)))

        predictions = {}
        for file in files:
            xml_path = os.path.join(os.path.dirname(os.path.dirname(file)),
                                    os.path.join(LABELS_FOLDER, os.path.basename(file).split('.')[0] + '.xml'))

            if not os.path.exists(xml_path):
                continue

            tree = ElementTree.parse(xml_path)
            root = tree.getroot()

            name_field = root.findall('object/name')
            if len(name_field) != 1:
                continue

            file_prediction = self.predict(file)

            predictions[name_field[0].text] = {
                'prediction': file_prediction[0].obj,
                'score': str(file_prediction[2].data.max().tolist())
            }

        features = feature_layer.query(output_fields=[input_label_field]).features
        features_to_update = []
        for feature in features:
            if predictions.get(str(feature.attributes[input_label_field])):
                feature.attributes[output_label_field] = predictions.get(str(feature.attributes[input_label_field]))[
                    'prediction']
                if confidence_field:
                    feature.attributes[confidence_field] = predictions.get(str(feature.attributes[input_label_field]))[
                        'score']

                features_to_update.append(feature)

        return features_to_update

    def classify_features(
        self,
        feature_layer,
        labeled_tiles_directory,
        input_label_field,
        output_label_field,
        confidence_field=None,
        predict_function=None
    ):

        """
        Classifies the exported images and updates the feature layer with the prediction results in the ``output_label_field``.

        ====================================     ====================================================================
        **Argument**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        feature_layer                            Required. Feature Layer for classification.
        ------------------------------------     --------------------------------------------------------------------
        labeled_tiles_directory                  Required. Folder structure containing images and labels folder. The
                                                 chips should have been generated using the export training data tool in
                                                 the Labeled Tiles format, and the labels should contain the OBJECTIDs
                                                 of the features to be classified.
        ------------------------------------     --------------------------------------------------------------------
        input_label_field                        Required. Value field name which created the labeled tiles. This field
                                                 should contain the OBJECTIDs of the features to be classified. In case of
                                                 attachments this field is not used.
        ------------------------------------     --------------------------------------------------------------------
        output_label_field                       Required. Output column name to be added in the layer which contains predictions.
        ------------------------------------     --------------------------------------------------------------------
        confidence_field                         Optional. Output column name to be added in the layer which contains the confidence score.
        ------------------------------------     --------------------------------------------------------------------
        predict_function                         Optional. Used for calculation of final prediction result when each feature
                                                 has more than one attachment. The ``predict_function`` takes as input a list of tuples.
                                                 Each tuple has first element as the class predicted and second element is the confidence score.
                                                 The function should return the final tuple classifying the feature and its confidence
        ====================================     ====================================================================

        :return:
            Boolean : True/False if operation is sucessful

        """

        if predict_function is None:
            predict_function = _prediction_function

        if input_label_field and _check_esri_files(Path(labeled_tiles_directory)):
            features_to_update = self._classify_labeled_tiles(
                feature_layer,
                labeled_tiles_directory,
                input_label_field,
                output_label_field,
                confidence_field
            )
        elif os.path.exists(os.path.join(labeled_tiles_directory, 'mapping.txt')):
            json_file = os.path.join(labeled_tiles_directory, 'mapping.txt')
            with open(json_file) as file:
                feature_attachments_mapping = json.load(file)

            features_to_update = self._classify_attachments(
                feature_layer,
                labeled_tiles_directory,
                feature_attachments_mapping,
                feature_layer.properties['objectIdField'],
                output_label_field,
                confidence_field,
                predict_function
            )
        else:
            return False

        return FeatureClassifier._update_predictions_layer(
            feature_layer,
            features_to_update,
            output_label_field,
            confidence_field
        )

    def _categorize_feature_layer(
        self,
        feature_layer,
        raster,
        class_value_field,
        class_name_field,
        confidence_field,
        cell_size,
        coordinate_system,
        predict_function,
        batch_size,
        overwrite
    ):  
        #
        norm_mean = torch.tensor(imagenet_stats[0])
        norm_std = torch.tensor(imagenet_stats[1])

        # Check and create Fields 
        class_name_field_template = {
            "name": class_name_field.lower(),
            "type": "esriFieldTypeString",
            "alias": class_name_field,
            "sqlType": "sqlTypeOther",
            "length": 256,
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": ''
        }

        class_value_field_template = {
            "name": class_value_field.lower(),
            "type": "esriFieldTypeInteger",
            "alias": class_value_field,
            "sqlType": "sqlTypeOther",
            "nullable": True,
            "editable": True,
            "visible": True,
            "domain": None,
            "defaultValue": -999
        }

        to_delete = []
        to_create = []

        feature_layer_fields = { f['name'].lower():f for f in feature_layer.properties["fields"] }
        oid_field = feature_layer.properties['objectIdField']

        if class_value_field_template['name'] in feature_layer_fields:
            if overwrite:
                to_delete.append(feature_layer_fields[class_value_field_template['name']])
            else:
                e = Exception(f"The specified class_value_field '{class_value_field}' already exists, please specify a different name or set `overwrite=True`")
                raise(e)
        to_create.append(class_value_field_template)

        if class_name_field_template['name'] in feature_layer_fields:
            if overwrite:
                to_delete.append(feature_layer_fields[class_name_field_template['name']])
            else:
                e = Exception(f"The specified class_name_field '{class_name_field}' already exists, please specify a different name or set `overwrite=True`")
                raise(e)
        to_create.append(class_name_field_template)
        
        if confidence_field is not None:
            confidence_field_template = {
                "name": confidence_field.lower(),
                "type": "esriFieldTypeDouble",
                "alias": confidence_field,
                "sqlType": "sqlTypeDouble",
                "nullable": True,
                "editable": True,
                "visible": True,
                "domain": None,
                "defaultValue": -999
            }
            if confidence_field_template['name'] in feature_layer_fields:
                if overwrite:
                    to_delete.append(feature_layer_fields[confidence_field_template['name']])
                else:
                    e = Exception(f"The specified confidence_field '{confidence_field}' already exists, please specify a different name or set `overwrite=True`")
                    raise(e)
            to_create.append(confidence_field_template)

        feature_layer.manager.delete_from_definition({'fields': to_delete})
        feature_layer.manager.add_to_definition({'fields': to_create})

        # Get features for updation
        fields_to_update = [oid_field, class_value_field, class_name_field]
        if confidence_field is not None:
            fields_to_update.append(confidence_field)
        feature_layer_features = feature_layer.query(out_fields=",".join(fields_to_update), return_geometry=False).features
        update_store = {}

        if raster is not None:
            import arcpy

            #Arcpy Environment to export data
            arcpy.env.cellSize = cell_size
            arcpy.env.outputCoordinateSystem = coordinate_system
            arcpy.env.cartographicCoordinateSystem = coordinate_system

            feature_layer_url = feature_layer.url

            if feature_layer._token is not None:
                feature_layer_url = feature_layer_url + f"?token={feature_layer._token}"

            
            
            # Create Temporary ID field
            tempid_field = _tempid_field = 'f_fcuid'
            i = 1
            while tempid_field in feature_layer_fields:
                tempid_field = _tempid_field + str(i)
                i+=1
            arcpy.AddField_management(feature_layer_url, tempid_field, "LONG")
            #feature_layer.manager.add_to_definition({'fields': [tempid_field_template]})
            arcpy.CalculateField_management(feature_layer_url, tempid_field, f"{oid_field}", "SQL")

            temp_folder = arcpy.env.scratchFolder
            temp_datafldr = os.path.join(temp_folder, 'categorize_features_'+str(int(time.time())))
            result = arcpy.ia.ExportTrainingDataForDeepLearning(
                in_raster=raster,
                out_folder=temp_datafldr,
                in_class_data=feature_layer_url,
                image_chip_format="TIFF",
                tile_size_x=self._data.chip_size,
                tile_size_y=self._data.chip_size,
                stride_x=0,
                stride_y=0,
                output_nofeature_tiles="ALL_TILES",
                metadata_format="Labeled_Tiles",
                start_index=0,
                class_value_field=tempid_field,
                buffer_radius=0,
                in_mask_polygons=None,
                rotation_angle=0
            )
            # cleanup
            arcpy.DeleteField_management(feature_layer_url, [ tempid_field ])

            image_list = ImageList.from_folder(os.path.join(temp_datafldr, 'images'))
            def get_id(imagepath):
                with open(os.path.join(temp_datafldr, 'labels', os.path.basename(imagepath)[:-3]+'xml')) as f:
                    return(int(f.read().split('<name>')[1].split('<')[0]))

            for i in range(0, len(image_list), batch_size):
                # Get Temporary Ids
                tempids = [ get_id(f) for f in image_list.items[i:i+batch_size] ]
                
                # Get Image batch
                image_batch = torch.stack([ im.data for im in image_list[i:i+batch_size] ])
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                
                # Get Predications
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                
                # push prediction to store
                for ui, oid in enumerate(tempids):
                    classvalue = self._data.classes[predicted_classes[ui]]
                    update_store[oid] = {
                        oid_field: oid,
                        class_value_field: classvalue,
                        class_name_field:  self._data.class_mapping[classvalue]
                    } 
                    if confidence_field is not None:
                        update_store[oid][confidence_field] = predictions_conf[ui]

            # Cleanup
            arcpy.Delete_management(temp_datafldr)
            shutil.rmtree(temp_datafldr, ignore_errors=True)

        else:
            out_folder = tempfile.TemporaryDirectory().name
            os.mkdir(out_folder)
            feature_layer.export_attachments(out_folder)
            with open(os.path.join(out_folder, 'mapping.txt')) as file:
                feature_attachments_mapping = json.load(file)
                images_store = []
                for oid in feature_attachments_mapping:
                    for im in feature_attachments_mapping[oid]:
                        images_store.append({
                            'oid': oid,
                            'im': os.path.join(out_folder,im)
                        })
            update_store_scratch = {}
            for i in range(0, len(images_store), batch_size):
                rel_objectids = []
                image_batch = []
                for r in images_store[i:i+batch_size]:
                    im = open_image(r['im']) # Read Bytes
                    im = im.resize(self._data.chip_size) # Resize
                    image_batch.append(im.data) # Convert to tensor
                    rel_objectids.append(int(r['oid']))
                image_batch = torch.stack(image_batch)
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                # Get Predictions and save to scratch
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                for ai, oid in enumerate(rel_objectids):
                    if update_store_scratch.get(oid) is None:
                        update_store_scratch[oid] = []
                    update_store_scratch[oid].append([predicted_classes[ai], predictions_conf[ai]])
            # Prepare final updated features
            for oid in update_store_scratch:
                max_prediction_class, max_prediction_value = predict_function(update_store_scratch[oid])
                if max_prediction_class is not None:
                    classvalue = self._data.classes[max_prediction_class]
                    classname = self._data.class_mapping[classvalue]
                else:
                    classvalue = None
                    classname = None                
                update_store[oid] = {
                    oid_field: oid,
                    class_value_field: classvalue,
                    class_name_field:  classname
                } 
                if confidence_field is not None:
                    update_store[oid][confidence_field] = max_prediction_value

        # Update Features
        features_to_update = []
        for feat in feature_layer_features:
            if update_store.get(feat.attributes[oid_field]) is not None:
                updated_attributes = update_store[feat.attributes[oid_field]]
                for f in fields_to_update:
                    feat.attributes[f] = updated_attributes[f]
                features_to_update.append(feat)
        step = 100
        for si in range(0, len(features_to_update), step):
            feature_batch = features_to_update[si:si+step]
            response = feature_layer.edit_features(updates=feature_batch)
            for resp in response.get('updateResults', []):
                if resp.get('success', False):
                    continue
                warnings.warn(f"Something went wrong for data {resp}")
            time.sleep(2)


    def _categorize_feature_class(
        self,
        feature_class,
        raster,
        class_value_field,
        class_name_field,
        confidence_field,
        cell_size,
        coordinate_system,
        predict_function,
        batch_size,
        overwrite
    ):
        import arcpy
        arcpy.env.overwriteOutput = overwrite

        if batch_size is None:
            batch_size  = self._data.batch_size

        if predict_function is None:
            predict_function = _prediction_function
        
        norm_mean = torch.tensor(imagenet_stats[0])
        norm_std = torch.tensor(imagenet_stats[1])
        
        fcdesc = arcpy.Describe(feature_class)
        oid_field = fcdesc.OIDFieldName
        if not (fcdesc.dataType == 'FeatureClass' and fcdesc.shapeType == 'Polygon'):
            e = Exception(f"The specified FeatureClass at '{feature_class}' is not valid, it should be Polygon FeatureClass")
            raise(e)
        fields = arcpy.ListFields(feature_class)
        field_names = [f.name for f in fields]
        if class_value_field in field_names:
            if not overwrite:
                e = Exception(f"The specified class_value_field '{class_value_field}' already exists in the target FeatureClass, please specify a different name or set `overwrite=True`")
                raise(e)
        arcpy.DeleteField_management(feature_class, 
                                [ class_value_field ])
        arcpy.AddField_management(feature_class, class_value_field, "LONG")
            
        if class_name_field in field_names:
            if not overwrite:
                e = Exception(f"The specified class_name_field '{class_name_field}' already exists in the target FeatureClass, please specify a different name or set `overwrite=True`")
                raise(e)
        arcpy.DeleteField_management(feature_class, 
                                [ class_name_field ])
        arcpy.AddField_management(feature_class, class_name_field, "TEXT")

        if confidence_field is not None:
            if confidence_field in field_names:
                if not overwrite:
                    e = Exception(f"The specified confidence_field '{confidence_field}' already exists in the target FeatureClass, please specify a different name or set `overwrite=True`")
                    raise(e)
            arcpy.DeleteField_management(feature_class, 
                                    [ confidence_field ])
            arcpy.AddField_management(feature_class, confidence_field, "DOUBLE")

        if raster is not None:
            #Arcpy Environment to export data
            arcpy.env.cellSize = cell_size
            arcpy.env.outputCoordinateSystem = coordinate_system
            arcpy.env.cartographicCoordinateSystem = coordinate_system

            tempid_field = _tempid_field = 'f_fcuid'
            i = 1
            while tempid_field in field_names:
                tempid_field = _tempid_field + str(i)
                i+=1
            arcpy.AddField_management(feature_class, tempid_field, "LONG")
            arcpy.CalculateField_management(feature_class, tempid_field, f"!{oid_field}!")

            temp_folder = arcpy.env.scratchFolder
            temp_datafldr = os.path.join(temp_folder, 'categorize_features_'+str(int(time.time())))
            result = arcpy.ia.ExportTrainingDataForDeepLearning(
                in_raster=raster,
                out_folder=temp_datafldr,
                in_class_data=feature_class,
                image_chip_format="TIFF",
                tile_size_x=self._data.chip_size,
                tile_size_y=self._data.chip_size,
                stride_x=0,
                stride_y=0,
                output_nofeature_tiles="ALL_TILES",
                metadata_format="Labeled_Tiles",
                start_index=0,
                class_value_field=tempid_field,
                buffer_radius=0,
                in_mask_polygons=None,
                rotation_angle=0
            )
            # cleanup
            arcpy.DeleteField_management(feature_class, [ tempid_field ])
            image_list = ImageList.from_folder(os.path.join(temp_datafldr, 'images'))
            def get_id(imagepath):
                with open(os.path.join(temp_datafldr, 'labels', os.path.basename(imagepath)[:-3]+'xml')) as f:
                    return(int(f.read().split('<name>')[1].split('<')[0]))

            for i in range(0, len(image_list), batch_size):
                # Get Temporary Ids
                tempids =[ get_id(f) for f in image_list.items[i:i+batch_size] ]
                
                # Get Image batch
                image_batch = torch.stack([ im.data for im in image_list[i:i+batch_size] ])
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                
                # Get Predications
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                
                # Update Feature Class
                where_clause = f"{oid_field} IN ({','.join(str(e) for e in tempids)})"
                update_cursor = arcpy.UpdateCursor(
                    feature_class,
                    where_clause=where_clause,
                    sort_fields=f"{oid_field} A"
                )
                for row in update_cursor:
                    row_tempid = row.getValue(oid_field)
                    ui = tempids.index(row_tempid)
                    classvalue = self._data.classes[predicted_classes[ui]]
                    row.setValue(class_value_field, classvalue)
                    row.setValue(class_name_field, self._data.class_mapping[classvalue])
                    if confidence_field is not None:
                        row.setValue(confidence_field, predictions_conf[ui])
                    update_cursor.updateRow(row)

                # Remove Locks
                del row
                del update_cursor

            # Cleanup
            arcpy.Delete_management(temp_datafldr)
            shutil.rmtree(temp_datafldr, ignore_errors=True)

        else:
            feature_class_attach = feature_class+'__ATTACH'
            nrows = arcpy.GetCount_management(feature_class_attach)[0]
            store={}
            for i in range(0, int(nrows), batch_size):
                attachment_ids = []
                rel_objectids = []
                image_batch = []
                
                # Get Image Batch
                with arcpy.da.SearchCursor(feature_class_attach, [ 'ATTACHMENTID', 'REL_OBJECTID', 'DATA' ]) as search_cursor:
                    for c, item in enumerate(search_cursor):
                        if c >= i and c < i+batch_size :
                            attachment_ids.append(item[0])
                            rel_objectids.append(item[1])
                            attachment = item[-1]
                            im = open_image(io.BytesIO(attachment.tobytes())) # Read Bytes
                            im = im.resize(self._data.chip_size) # Resize
                            image_batch.append(im.data) # Convert to tensor
                            del item
                            del attachment
                            #del im
                image_batch = torch.stack(image_batch)
                image_batch = normalize(image_batch, mean=norm_mean, std=norm_std)
                
                # Get Predictions and save to store
                predicted_classes, predictions_conf = self._predict_batch(image_batch)
                for ai in range(len(attachment_ids)):
                    if store.get(rel_objectids[ai]) is None:
                        store[rel_objectids[ai]] = []
                    store[rel_objectids[ai]].append([predicted_classes[ai], predictions_conf[ai]])
                
            # Update Feature Class
            update_cursor = arcpy.UpdateCursor(feature_class)
            for row in update_cursor:
                row_oid = row.getValue(oid_field)
                max_prediction_class, max_prediction_value = predict_function(store[row_oid])
                if max_prediction_class is not None:
                    classvalue = self._data.classes[max_prediction_class]
                    classname = self._data.class_mapping[classvalue]
                else:
                    classvalue = None
                    classname = None
                row.setValue(class_value_field, classvalue)
                row.setValue(class_name_field, classname)
                if confidence_field is not None:
                    row.setValue(confidence_field, max_prediction_value)
                update_cursor.updateRow(row)

            # Remove Locks
            del row
            del update_cursor
        return True

    def categorize_features(
        self,
        feature_layer,
        raster=None,
        class_value_field='class_val',
        class_name_field='prediction',
        confidence_field="confidence",
        cell_size=1,
        coordinate_system=3857,
        predict_function=None,
        batch_size=64,
        overwrite=False 
    ):
        """
        Categorizes each feature by classifying its attachments or an image of its geographical area (using the provided Imagery Layer)
        and updates the feature layer with the prediction results in the ``output_label_field``.

        ====================================     ====================================================================
        **Argument**                             **Description**
        ------------------------------------     --------------------------------------------------------------------
        feature_layer                            Required. Feature Layer or path of local feature class for classification with read, write, edit permissions.
        ------------------------------------     --------------------------------------------------------------------
        raster                                   Optional. Imagery layer or path of local raster to be used for exporting image chips. (Requires arcpy)
        ------------------------------------     --------------------------------------------------------------------
        class_value_field                        Required string. Output field to be added in the layer, containing class value of predictions.
        ------------------------------------     --------------------------------------------------------------------
        class_name_field                         Required string. Output field to be added in the layer, containing class name of predictions.
        ------------------------------------     --------------------------------------------------------------------
        confidence_field                         Optional string. Output column name to be added in the layer which contains the confidence score.
        ------------------------------------     --------------------------------------------------------------------
        cell_size                                Optional float. Cell size to be used for exporting the image chips.
        ------------------------------------     --------------------------------------------------------------------
        coordinate_system                        Optional. Cartographic Coordinate System to be used for exporting the image chips.
        ------------------------------------     --------------------------------------------------------------------
        predict_function                         Optional list of tuples. Used for calculation of final prediction result when each feature
                                                 has more than one attachment. The ``predict_function`` takes as input a list of tuples.
                                                 Each tuple has first element as the class predicted and second element is the confidence score.
                                                 The function should return the final tuple classifying the feature and its confidence.
        ------------------------------------     --------------------------------------------------------------------
        batch_size                               Optional integer. The no of images or tiles to process in a single go. 
                                                 
                                                 The default value is 64.
        ------------------------------------     --------------------------------------------------------------------
        overwrite                                Optional boolean. If set to True the output fields will be overwritten by new values. 

                                                 The default value is False.
        ====================================     ====================================================================

        :return:
            Boolean : True if operation is successful, False otherwise

        """        

        import arcgis
        from arcgis.features import FeatureLayer
        from arcgis.raster import ImageryLayer
        from arcgis.gis import Item
        
        if predict_function is None:
            predict_function = _prediction_function

        if isinstance(raster, str):
            if 'http' in raster:
                raster = ImageryLayer(raster, gis=arcgis.env.active_gis)
        if isinstance(raster, ImageryLayer):
            raster_url = raster.url
            if raster._token is not None:
                raster_url = raster_url + f"?token={raster._token}" 
            raster = raster_url
        elif isinstance(raster, Item):  # handles Mapserver
            raster_url = raster.url
            if raster.layers[0]._token is not None:
                raster_url = raster_url + f"?token={raster.layers[0]._token}" 
            raster = raster_url

        if isinstance(feature_layer, str):
            if 'http' in feature_layer:
                feature_layer = FeatureLayer(feature_layer, gis=arcgis.env.active_gis)
            else:
                return self._categorize_feature_class(
                    feature_class=feature_layer,
                    raster=raster,
                    class_value_field=class_value_field,
                    class_name_field=class_name_field,
                    confidence_field=confidence_field,
                    cell_size=cell_size,
                    coordinate_system=coordinate_system,
                    predict_function=predict_function,
                    batch_size=batch_size,
                    overwrite=overwrite
                )

        if isinstance(feature_layer, FeatureLayer):
            return self._categorize_feature_layer(
                feature_layer=feature_layer,
                raster=raster,
                class_value_field=class_value_field,
                class_name_field=class_name_field,
                confidence_field=confidence_field,
                cell_size=cell_size,
                coordinate_system=coordinate_system,
                predict_function=predict_function,
                batch_size=batch_size,
                overwrite=overwrite
            )
        else:
            e = Exception("Could not understand layer type")
            raise(e)
            



try:
    import spacy
    from spacy.gold import offsets_from_biluo_tags as _offsets_from_biluo_tags
    from spacy.gold import iob_to_biluo as _iob_to_biluo
    import pandas as pd
    HAS_SPACY = True
except:
    HAS_SPACY = False
from pathlib import Path
import json,random,os,tempfile,logging

__all__=["_from_bio_tags","_from_json","ner_prepare_data","_create_zip"]

def _raise_spacy_import_error():
    raise Exception('This module requires pandas and spacy version 2.1.8. Install it using \"pip install pandas spacy==2.1.8\"')

def _create_zip(zipname, path):
    import shutil

    if os.path.exists(os.path.join(path, zipname) + '.zip'):
        os.remove(os.path.join(path, zipname) + '.zip')
        
    temp_dir = tempfile.TemporaryDirectory().name    
    zip_file = shutil.make_archive(os.path.join(temp_dir, zipname), 'zip', path)
    
    shutil.move(zip_file, path)


def _from_bio_tags(tokens_collection, tags_collection):
    """
    Converts training data from ``BIO`` format to spacy offsets.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    tokens_collection       Required [list]. List of token lists
                            Example: [[This,is,a,test],[This,is,a,test1]]
    ---------------------   -------------------------------------------
    tags_collection         Required [list]. List of tag lists
                            Example: [[B-tagname,O,O,O],[B-tagname,I-tagname,O,O]]
    =====================   ===========================================
    """


    nlp=spacy.blank('en')
    
    train_data = [] 
    for tags, tokens in zip(tags_collection, tokens_collection):
        
        try:
            tags = _iob_to_biluo(tags)

            doc = spacy.tokens.doc.Doc(
            nlp.vocab, words = tokens, spaces = [True]*(len(tokens)-1)+[False])
            # run the standard pipeline against it
            for name, proc in nlp.pipeline:
                doc = proc(doc)
            
            text=' '.join(tokens)
            tags = _offsets_from_biluo_tags(doc, tags)
            train_data.append((text,{'entities':tags}))
        except:
            pass
        

    return train_data
        


def _from_json(path, text_key='text', offset_key='labels'):
    """
    Converts training data from JSON format to spacy offsets.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    text_key                Optional:str='text. Json key under which text is available
    ---------------------   -------------------------------------------
    offset_key              Optional:str='labels. Json key under which offsets are available
    =====================   ===========================================
    json-schema:
    ----------
    {"id": 1, "text": "EU rejects ...", "labels": [[0,2,"ORG"], [11,17, "MISC"], [34,41,"ORG"]]}
    {"id": 2, "text": "Peter Blackburn", "labels": [[0, 15, "PERSON"]]}
    {"id": 3, "text": "President Obama", "labels": [[10, 15, "PERSON"]]}
    ----------
    returns: A json file that can be consumed by ner_databunch.
    """
    
    train_data = []
    with open(path,'r', encoding='UTF-8') as f:
        data_list = f.readlines()
    for i, item in enumerate(data_list):
        try:
            train_data.append((json.loads(item).get(text_key), {'entities':json.loads(item).get(offset_key)}))
        except:
            pass
        
    return train_data

def ner_prepare_data(dataset_type, path, class_mapping=None, val_split_pct=0.1):

    """
    Prepares a data object

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    dataset_type            Required string. ['ner_json', 'BIO', 'LBIOU']
    ---------------------   -------------------------------------------
    address_tag             Optional dict. Address field/tag name 
                            in the training data.
    val_split_pct           Optional Float. Percentage of training data to keep
                            as validation. The default value is 0.1.
    =====================   ===========================================
    returns: A list [text,{entities},text,{entities}] that can be ingested by ``EntityRecognizer``.
    """
    import spacy
    v_list=spacy.__version__.split('.')
    version=sum([int(j)*10**(2*i) for i,j in enumerate(v_list[::-1])])
    if version<20108: #checking spacy version
        return logging.error(f'Entity recognition model needs spacy version 2.1.8 or higher. Your current spacy version is {spacy.__version__}, please update using \'pip install')

    if not HAS_SPACY:
        _raise_spacy_import_error()
    path=Path(path)
    if class_mapping:
        address_tag=class_mapping.get('address_tag')
    
    else:
        address_tag='Address'

    if dataset_type == 'ner_json':
        train_data = _from_json(path=path)
        path=path.parent
    elif dataset_type == 'BIO':
        tags_collection = []
        tokens_collection = []
        tags_df = pd.read_csv(path/'tags.csv')
        tokens_df = pd.read_csv(path/'tokens.csv')
        
        for i,tags in tags_df.iterrows():
            tags_collection.append(list(tags.dropna()))
        
        for i,tokens in tokens_df.iterrows():
            tokens_collection.append(list(tokens.dropna()))

        train_data = _from_bio_tags(tags_collection=tags_collection, tokens_collection=tokens_collection)
    elif dataset_type == 'LBIOU':

        tags_collection = []
        tokens_collection = []
        tags_df = pd.read_csv(path/'tags.csv')
        tokens_df = pd.read_csv(path/'tokens.csv')
        
        for i,tags in tags_df.iterrows():
            tags_collection.append(list(tags.dropna()))
        
        for i,tokens in tokens_df.iterrows():
            tokens_collection.append(list(tokens.dropna()))
        train_data = _offsets_from_biluo_tags(tags=tags_collection, tokens=tokens_collection)
    # return train_data
    data=DatabunchNER(train_data, val_split_pct=val_split_pct, address_tag=address_tag, test_ds=None)
    data.path=path
    return data

class _NERItemlist():
    """
    Creates a dataset to store data within ``ner_databunch`` object.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    bs                      Batch size. 
    ---------------------   -------------------------------------------
    data                    Required:DatabunchNER. 
    =====================   ===========================================

    :returns: dataset.
    """

    
    def __init__(self, bs, data):
        self.bs = bs
        self.entities = ['TEXT']+list({i[2] for i in pd.concat([pd.Series(i['entities']) for i in [o[1] for o in data]])}) ##Extracting all the unique entity names from input json
        self.data = data
        self.x = [o[0] for o in data]
        self.y = [o[1] for o in data]

    def __getitem__(self, i):
        return self.data[i]

    def __len__(self):
        return len(self.data)

    def _random_batch(self, data):
        res = []
        for j in range(self.bs): 
            res.append(random.choice(data)) 
        return res    
    
    def _entities_to_dataframe(self, item):
        """
        This function is used to create pandas dataframe from training input data json.
        """
        text = item[0]
        df = pd.DataFrame(item[1].get('entities'))

        out_dict = {}
        for x in df[2].unique(): out_dict[x] = df[df[2] == x][[0, 1]].values.tolist()

        out = {}
        out['text'] = text
        for key in out_dict.keys():
            
            for tpl in out_dict.get(key):
                if out.get(key) == None:
                    out[key] = []
                out[key].append(text[tpl[0]:tpl[1]])
        return pd.Series(out)


    def show_batch(self):
        """
        This function shows a batch from the _NERItemlist.
        """
        data = self._random_batch(self.data)
        lst = []
        for item in data:
            lst.append(self._entities_to_dataframe(item))
        batch_df = pd.concat(lst,axis=1,sort=True).T
        batch_df

        return batch_df.fillna('')

    

class DatabunchNER():


    """
    Creates a databunch object.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    ds                      Required data list returned from _ner_prepare_data(). 
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
                            The default Value is 0.1.
    ---------------------   -------------------------------------------
    bs                      Optional integer. Batch size
                            The default value is 5.                       
    =====================   ===========================================

    :returns: dataset
    """

    def __init__(self, ds, val_split_pct, bs=5, test_ds=None,address_tag=None):
        random.shuffle(ds)
        self.train_ds = _NERItemlist(bs,data = ds[:int(len(ds)*(1-val_split_pct))]) #creating an _NERItemlist with training dataset
        self.val_ds = _NERItemlist(bs,data = ds[int(len(ds)*(1-val_split_pct)):]) #creating an _NERItemlist with validation dataset
        self.entities=list(set(self.train_ds.entities).union(set(self.val_ds.entities)))
        self._address_tag=address_tag
        self._has_address=True
        if self._address_tag not in self.entities:
            self._has_address=False
            return logging.warning("No Address tag found in your data.\n\
                1. If your data has an address field, pass your address field name as address tag in class mapping \n\
                e.g. - data=prepare_data(dataset_type=ds_type,path=training_data_folder,\n\t\t\t\
                    class_mapping={address_tag:address_field_name})\n\
                2. Else no action is required, if your data does not have any address information.")

    def show_batch(self):
        return self.train_ds.show_batch()
    


import os
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap

########################################################################
class SystemManager(object):
    """
    The System resource is a collection of server-wide resources in your
    ArcGIS Notebook Server site. Within this resource, you can access
    information and perform operations pertaining to licenses, Web
    Adaptors, containers, server properties, directories, Jobs, and the
    configuration store.
    """
    _url = None
    _con = None
    _gis = None
    _dir = None
    _wam = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            url = self._url + "/properties"
            params = {'f': 'json'}
            res = self._con.get(url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<SystemManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<SystemManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """
        ArcGIS Notebook Server has configuration properties that govern
        some of its intricate behavior. This resource is a container for
        these properties. The properties are available to all server
        objects and extensions through the server environment interface.


        The available properties are as follows:

            + WebContextURL - Defines the web front-end as seen by your users. Example: https://mydomain.com/gis
            + maxContainersPerNode - The default maximum number of containers that can be opened on a notebook server machine assuming the machine has the necessary CPU/Memory resources to support the containers.
            + idleNotebookThreshold - Specifies the time (in minutes) after which idle notebooks are closed automatically.
            + containerCreatedThreshold - Specifies the time (in minutes) after which an empty container is closed automatically.
            + webSocketSize - Specifies the amount of memory (in MB) available to ArcGIS Notebooks for WebSocket communication

        :returns: PropertyMap
        """
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    @properties.setter
    def properties(self, value):
        """
        Sets the ArcGIS Notebook Server has configuration properties that govern
        some of its intricate behavior. This resource is a container for
        these properties. The properties are available to all server
        objects and extensions through the server environment interface.


        The available properties are as follows:

            + WebContextURL - Defines the web front-end as seen by your users. Example: https://mydomain.com/gis
            + maxContainersPerNode - The default maximum number of containers that can be opened on a notebook server machine assuming the machine has the necessary CPU/Memory resources to support the containers.
            + idleNotebookThreshold - Specifies the time (in minutes) after which idle notebooks are closed automatically.
            + containerCreatedThreshold - Specifies the time (in minutes) after which an empty container is closed automatically.
            + webSocketSize - Specifies the amount of memory (in MB) available to ArcGIS Notebooks for WebSocket communication

        :returns: PropertyMap
        """
        properties: {
          "dockerConnectionPort": 2375,
          "webSocketSize": 16,
          "maxContainersPerNode": 20,
          "containersStartPort": 30001,
          "containersStopPort": 31000,
          "idleNotebookThreshold": 1440,
          "containerCreatedThreshold": 60,
          "dockerConnectionHost": "localhost"
        }
        props = {}
        url = self._url + "/properties/update"
        params = {'f' : 'json',
                  'properties' : {}}
        current = dict(self.properties)
        for k in current.keys():
            if k in value:
                params['properties'][k] = value[k]
            else:
                params['properties'][k] = current[k]
        res = self._con.post(url, params)
        if not 'status' in res:
            raise Exception(res)
    #----------------------------------------------------------------------
    @property
    def containers(self):
        """
        Returns a list of active containers.

        :returns: List of Container classes
        """
        container = []
        url = self._url + "/containers"
        params = {'f' : 'json'}
        res = self._con.get(url, params)
        if "containers" in res:
            for c in res["containers"]:
                cid = c['id']
                curl = self._url + "/containers/{cid}".format(cid=cid)
                container.append(Container(url=curl, gis=self._gis))
        return container

    #----------------------------------------------------------------------
    @property
    def licenses(self):
        """
        Gets the license resource list.  The licenses resource lists the
        current license level of ArcGIS Notebook Sever and all authorized
        extensions. Contact Esri Customer Service if you have questions
        about license levels or expiration properties.
        """
        url = self._url + "/licenses"
        params = {
            "f" : "json"
        }
        return self._con.get(url,
                             params)
    #----------------------------------------------------------------------
    @property
    def web_adaptors(self):
        """
        returns a list of web adapters

        :return: List
        """
        if self._wam is None:
            url = self._url + "/webadaptors"
            self._wam = WebAdaptorManager(url=url, gis=self._gis)
        return self._wam
    #----------------------------------------------------------------------
    @property
    def jobs(self):
        """
        This resource is a collection of all the administrative jobs
        (asynchronous operations) created within your site. When operations
        that support asynchronous execution are run, ArcGIS Notebook Server
        creates a new job entry that can be queried for its current status
        and messages.

        :returns: list

        """
        url = self._url + "/jobs"
        params = {'f' : 'json'}
        res = self._con.get(url, params)
        if "asyncJobs" in res:
            return res["asyncJobs"]
        return res
    #----------------------------------------------------------------------
    @property
    def directories(self):
        """Provides access to registering directories"""
        if self._dir is None:
            url = self._url + "/directories"
            self._dir = DirectoryManager(url=url, gis=self._gis)
        return self._dir
    #----------------------------------------------------------------------
    def job_details(self, job_id):
        """
        A job represents the asynchronous execution of an operation in
        ArcGIS Notebook Server. You can acquire progress information by
        periodically querying the job.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        job_id                 Required String. The unique identifier of the job.
        ==================     ====================================================================

        :returns: dict

        """
        url = self._url + "/jobs/{jid}".format(jid=job_id)
        params = {'f':'json'}
        return self._con.get(url, params)
    #----------------------------------------------------------------------
    @property
    def config_store(self):
        """
        The configuration store maintains configurations for ArcGIS Notebook
        Server. Typical configurations include all the resources such as
        machines and security rules that are required to power the site. In
        a way, the configuration store is a physical representation of a site.

        Every ArcGIS Notebook Server machine, when it joins the site, is
        provided with a connection to the configuration store and it can
        thereafter participate in the management of the site. You can change
        the store's properties during runtime using the edit operation.

        The Administrator API that runs on every server machine is capable
        of reading and writing to the store. As a result, the store must be
        accessible to every server machine within the site. The default
        implementation is built on top of a file system and stores all the
        configurations in a hierarchy of folders and files.

        :returns: dict

        """
        url = self._url + "/configStore"
        params = {'f' : 'json'}
        return self._con.get(url, params)

########################################################################
class DirectoryManager(object):
    """
    A manages and maintains a collection of all server directories.
    """
    _url = None
    _con = None
    _gis = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            url = self._url
            params = {'f': 'json'}
            res = self._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<DirectoryManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<DirectoryManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    def list(self):
        """
        returns the current registered directories

        :return: List

        """
        self._properties = None
        val = dict(self.properties)
        return val['directories']
    #----------------------------------------------------------------------
    def register(self, name, path, directory_type):
        """
        This operation registers a new data directory from your local
        machine with the ArcGIS Notebook Server site. Registering a local
        folder as a data directory allows your notebook authors to work with
        files in the folder.

        ==================     ====================================================================
        **Parameter**          **Description**
        ------------------     --------------------------------------------------------------------
        name	               The name of the directory.
        ------------------     --------------------------------------------------------------------
        path	               The full path to the directory on your machine.
        ------------------     --------------------------------------------------------------------
        directory_type	       The type of directory. Values: DATA | WORKSPACE | OUTPUT
        ==================     ====================================================================

        :returns: boolean

        """
        params = {
            'f' : 'json',
            'name' : name,
            'path' : path,
            'type' : directory_type
        }
        url = self._url + "/register"
        res = self._con.post(url, params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def unregister(self, directory_id):
        """
        This operation unregisters an existing directory from the ArcGIS
        Notebook Server site.

        ==================     ====================================================================
        **Parameter**          **Description**
        ------------------     --------------------------------------------------------------------
        directory_id           Required String.  The directory ID to remove.
        ==================     ====================================================================

        :returns: boolean

        """
        params = {'f' : 'json'}
        url = self._url + "/{uid}/unregister" .format(uid=directory_id)
        res = self._con.post(url, params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
########################################################################
class WebAdaptorManager(object):
    """
    Manages and configures web adaptors for the ArcGIS Notebook Server.
    """
    _url = None
    _con = None
    _gis = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<WebAdapterManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<WebAdapterManager @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    def register(self,
                 name,
                 ip,
                 webadapter_url,
                 http_port,
                 https_port,
                 description=""):
        """
        Registers a new web adapter.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required String. The name of the web adapter
        ------------------     --------------------------------------------------------------------
        ip                     Required String. The IP of the web adapter.
        ------------------     --------------------------------------------------------------------
        webadapter_url         Required String. The URI endpoint of the web adpater.
        ------------------     --------------------------------------------------------------------
        http_port              Required Integer. The port number of the web adapter
        ------------------     --------------------------------------------------------------------
        https_port             Required Integer. The secure port of the web adapter.
        ------------------     --------------------------------------------------------------------
        description            Optional String. The optional web adapter description.
        ==================     ====================================================================

        :returns: Boolean

        """
        params = {
            "f" : "json",
            "machineName" : name,
            "machineIP" : ip,
            "webAdaptorURL" : webadaptor_url,
            "description" : description,
            "httpPort" : http_port,
            "httpsPort" : https_port
        }
        url = self._url + "/register"
        res = self._con.post(url, params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    @property
    def config(self):
        """
        Gets the Web Adaptors configuration which is a resource of all the
        configuration parameters shared across all the Web Adaptors in the
        site. Most importantly, this resource lists the shared key that is
        used by all the Web Adaptors to encrypt key data bits for the
        incoming requests to the server.
        """
        url = self._url + "/config"
        params = {
            "f" : "json"
        }
        return self._con.get(url,
                             params)
    #----------------------------------------------------------------------
    @config.setter
    def config(self, config):
        """
        This is a property that allows for the retreival and manipulation of web adaptors.

        You can use this operation to change the Web Adaptor configuration
        and the sharedkey attribute. The sharedkey attribute must be present
        in the request.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        config                 Required dict. The configuration items to be updated for this web
                               adaptor. Always include the web adaptor's sharedkey attribute.
        ==================     ====================================================================

        :return:
            A boolean indicating success (True), else a Python dictionary containing an error message.

        """
        url = self._url + "/config/update"
        params = {
            "f" : "json",
            "webAdaptorConfig" : config
        }
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def list(self):
        """
        Returns all registered Web Adapters

        :return: List
        """
        url = self._url
        params = {'f' : 'json'}
        res = self._con.get(url, params)
        if "webAdaptors" in res:
            return [WebAdaptor(self._url + "/{wa}".format(wa=wa['id']),
                               gis=self._gis) for wa in res["webAdaptors"]]
        return res
########################################################################
class WebAdaptor(object):
    """
    This resource provides information about the ArcGIS Web Adaptor
    configured with your ArcGIS Notebook Server site. ArcGIS Web Adaptor
    is a web application that runs in a front-end web server. One of the
    Web Adaptor's primary responsibilities is to forward HTTP requests
    from end users to ArcGIS Notebook Server in a round-robin fashion.
    The Web Adaptor acts a reverse proxy, providing the end users with
    an entry point into the system, hiding the server itself, and
    providing some degree of immunity from back-end failures.

    The front-end web server could authenticate incoming requests against
    your enterprise identity stores and provide specific authentication
    schemes like Integrated Windows Authentication (IWA), HTTP Basic or
    Digest.

    Most importantly, ArcGIS Web Adaptor provides your end users with a
    well-defined entry point into your system without exposing the internal
    details of your server site. ArcGIS Notebook Server will trust requests
    being forwarded by ArcGIS Web Adaptor and will not challenge the user
    for any credentials. However, the authorization of the request (by
    looking up roles and permissions) is still enforced by the server site.

    ArcGIS Notebooks use the WebSocket protocol for communication. You can
    update the maximum size of the file sent using WebSocket by updating your
    site's webSocketMaxHeapSize property.
    """
    _url = None
    _con = None
    _gis = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<WebAdapter @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<WebAdapter @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def unregister(self):
        """
        Unregisters a WebAdapter for the Notebook Server
        :returns: boolean
        """
        url = self._url + "/unregister"
        params = {'f':'json'}
        res = self._con.post(url, params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
########################################################################
class Container(object):
    """
    This represents a single hosted notebook container.
    """
    _url = None
    _con = None
    _gis = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<Container @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<Container @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    @property
    def sessions(self):
        """
        When an ArcGIS Notebook is opened in a container, a computational
        engine called a kernel is launched; the kernel runs while the notebook
        is active and executes all the work done by the notebook. This
        resource tracks the active kernels running in the specified container
        to provide information about current notebook sessions.


        :returns: list of dict

        ==================     ====================================================================
        **Response**           **Description**
        ------------------     --------------------------------------------------------------------
        ID	               The unique ID string for the container session.
        ------------------     --------------------------------------------------------------------
        path	               The working path to the running ArcGIS Notebook, ending in .ipynb.
        ------------------     --------------------------------------------------------------------
        kernel	               Properties describing the kernel. They are as follows:

                                    + last_activity: The date and time of the most recent action performed by the kernel.
                                    + name: The name of the kernel. At 10.7, this value is python3.
                                    + id: The unique ID string for the kernel.
                                    + execution_state: Whether the kernel is currently executing an action or is idle.
                                    + connections: The number of users currently accessing the notebook.
        ==================     ====================================================================
        """
        url = self._url + "/sessions"
        params = {'f' : 'json'}
        res = self._con.get(url, params)
        if 'sessions' in res:
            return res['sessions']
        return res
    #----------------------------------------------------------------------
    def shutdown(self):
        """
        Terminates the current container

        :returns: boolean
        """
        url = self._url + "/terminateContainer"
        params = {'f' : 'json'}
        res = self._con.post(url, params)
        if 'status' in res:
            return res['status'] == 'success'
        return res






import sys
import copy
import json
from arcgis.gis import GIS, Item, User, Group
#--------------------------------------------------------------------------
def _search(gis, query, stype="content",
            max_items=100, bbox=None,
            categories=None, category_filter=None,
            start=1, sort_field="avgrating",
            sort_order="asc", count_fields=None,
            count_size=None, group_id=None, as_dict=False):
    """
    Generalized advanced search method.  This method allows for the query and
    searching of users, groups, group content, and general content.  This method
    allows for the control of the finer grained operations hidden by the 'search'
    method where a user can control and define how all information is returned.

    ================    ===============================================================
    **Argument**        **Description**
    ----------------    ---------------------------------------------------------------
    gis                 Required GIS. The connection object.
    ----------------    ---------------------------------------------------------------
    query               Required String.  The search query.
    ----------------    ---------------------------------------------------------------
    stype               Required String. The search type to find. This tells the
                        internal method what type of return object should be used.
                        Allowed values: `content`, `users`, `group_content`, and `groups`.
    ----------------    ---------------------------------------------------------------
    bbox                Optional String. This is the xmin,ymin,xmax,ymax bounding box to
                        limit the search in.  Items like documents do not have bounding
                        boxes and will not be included in the search.
    ----------------    ---------------------------------------------------------------
    categories          Optional String. A comma separated list of up to 8 org content
                        categories to search items. Exact full path of each category is
                        required, OR relationship between the categories specified.

                        Each request allows a maximum of 8 categories parameters with
                        AND relationship between the different categories parameters
                        called.
    ----------------    ---------------------------------------------------------------
    category_filters    Optional String. A comma separated list of up to 3 category
                        terms to search items that have matching categories. Up to 2
                        `category_filters` parameter are allowed per request. It can
                        not be used together with categories to search in a request.
    ----------------    ---------------------------------------------------------------
    start               Optional Int. The starting position to search from.  This is
                        only required if paging is needed.
    ----------------    ---------------------------------------------------------------
    sort_field          Optional String. Responses from the `search` operation can be
                        sorted on various fields. `avgrating` is the default.
    ----------------    ---------------------------------------------------------------
    sort_order          Optional String. The sequence into which a collection of
                        records are arranged after they have been sorted. The allowed
                        values are: asc for ascending and desc for descending.
    ----------------    ---------------------------------------------------------------
    count_fields        Optional String. A comma separated list of fields to count.
                        Maximum count fields allowed per request is 3. Supported count
                        fields: `tags`, `type`, `access`, `contentstatus`, and
                        `categories`.
    ----------------    ---------------------------------------------------------------
    count_size          Optional Int. The maximum number of field values to count for
                        each `count_fields`. The default value is None, and maximum size
                        allowed is 200.
    ----------------    ---------------------------------------------------------------
    group_id            Optional String. The unique `id` of the group to search for
                        content in. This is only used if `group_content` is used for
                        searching.
    ----------------    ---------------------------------------------------------------
    as_dict             Required Boolean. If True, the response comes back as a dictionary.
    ================    ===============================================================

    """
    if gis is None:
        import arcgis
        gis = arcgis.env.active_gis

    if max_items == -1:
        page_size = 100
    else:
        page_size = min(max_items, 100)
    items = []
    params = {
        "f" : "json",
        "q" : query,
        "start" : start,
        "num" : page_size,
        "sortField" : sort_field,
        "sortOrder" : sort_order
    }
    stype = str(stype).lower()
    if categories:
        params["categories"] = json.dumps(categories)
    if category_filter:
        params["categoryFilters"] = category_filter
    if count_fields:
        params['countFields'] = count_fields
    if count_size:
        params['countSize'] = count_size
    if bbox:
        if isinstance(bbox, (tuple, list)):
            bbox = ",".join([str(b) for b in bbox])
        params['bbox'] = bbox
    if stype in {"content", "item", "items"}:
        url = "{base}search".format(base=gis._portal.resturl)
    elif stype == "group_content" and group_id:
        url = "{base}content/groups/{gid}/search".format(base=gis._portal.resturl,
                                                        gid=group_id)
    elif stype == "group_content" and group_id is None:
        raise
    elif stype == "portal_users":
        allowed_keys = {'q', 'start', 'num', 'sortField',
                        'sortOrder', 'f', 'token'}
        for k in list(params.keys()):
            if not k in allowed_keys:
                del params[k]
            del k
        url = "{base}portals/self/users".format(base=gis._portal.resturl)
    elif stype in {"user", "groups", "users", "group"}:
        allowed_keys = {'q', 'start', 'num', 'sortField',
                        'sortOrder', 'f', 'token'}
        for k in list(params.keys()):
            if not k in allowed_keys:
                del params[k]
            del k
        if stype in ["user", "users"]:
            url = "{base}community/users".format(base=gis._portal.resturl)
        elif stype in {"group","groups"}:
            url = "{base}community/groups".format(base=gis._portal.resturl)
    count = 0
    res = gis._con.post(url, params)
    results = copy.deepcopy(res)
    count += int(res['num'])
    nextstart = int(res['nextStart'])
    while (count < max_items and max_items > 0) or (nextstart > 0 and max_items == -1):
        params['start'] = res['nextStart']
        if len(results['results']) >= max_items and max_items != -1:
            break
        res = gis._con.post(url, params)
        results['results'].extend(res['results'])
        count += int(res['num'])
        nextstart = int(res['nextStart'])
    if len(results['results']) > max_items and max_items != -1:
        results['results'] = results['results'][:max_items]
    ## Clean up Response
    ##
    results['results'] = _handle_response(as_dict=as_dict, gis=gis,
                                          res=results, stype=stype)
    results['nextStart'] = max_items + 1
    results['start'] = start
    return results
#--------------------------------------------------------------------------
def _handle_response(res, stype, gis, as_dict):
    """
    Handles returning the data in the proper format

    ================    ===============================================================
    **Argument**        **Description**
    ----------------    ---------------------------------------------------------------
    res                 Required Dict.  The response dictionary from the query call.
    ----------------    ---------------------------------------------------------------
    stype               Required String. The search type to find. This tells the
                        internal method what type of return object should be used.
    ----------------    ---------------------------------------------------------------
    gis                 Required GIS. The connection object.
    ----------------    ---------------------------------------------------------------
    as_dict             Required Boolean. If True, the response comes back as a dictionary.
    ================    ===============================================================

    :returns: List

    """
    if as_dict:
        return res['results']
    elif str(stype).lower() in {"content", "item", "items", "group_content"}:
        return [Item(itemid=r['id'], itemdict=r, gis=gis) for r in res['results']]
    elif str(stype).lower() in ["user", "users", "accounts", "account"]:
        return [User(gis=gis, username=r['username'], userdict=res) for r in res['results']]
    elif str(stype).lower() in ["groups", 'group']:
        return [Group(groupdict=r, groupid=r['id'], gis=gis) for r in res['results']]
    return res['results']



import pandas as pd
from arcgis.geometry import Geometry, Point, Polygon, Polyline, MultiPoint
from arcgis.features.geo._accessor import _is_geoenabled

def select(sdf, other):
    """
    Performs a select by location operation

    =========================    =========================================================
    **Argument**                 **Description**
    -------------------------    ---------------------------------------------------------
    sdf                          Required Spatially Enabled DataFrame. The dataframe to have the operation performed on.
    -------------------------    ---------------------------------------------------------
    other                        Required Spatially Enabled DataFrame or arcgis.Geometry.  This is the selecting data.
    =========================    =========================================================

    :returns: pd.DataFrame (Spatially enabled DataFrame)

    """
    ud = pd.Series([False] * len(sdf))
    if isinstance(other, (Point, Polygon, Polyline, MultiPoint)):
        sindex = sdf.spatial.sindex()
        q1 = sindex.intersect(bbox=other.extent)
        sub = sdf.iloc[q1]
        dj = sub[sdf.spatial.name].geom.disjoint(other) == False
        dj.index = sub.index
        ud = ud | dj
        return sdf[ud]
    elif _is_geoenabled(other):
        sindex = sdf.spatial.sindex()
        name = other.spatial.name
        for index, seg in other.iterrows():
            g = seg[name]
            q1 = sindex.intersect(bbox=g.extent)
            sub = sdf.iloc[q1]
            if len(sub) > 0:
                dj = sub[sdf.spatial.name].geom.disjoint(g) == False
                dj.index = sub.index
                ud = ud | dj
        return sdf[ud]
    else:
        raise ValueError(("Invalid input, please verify that `other` "
                          "is a Point, Polygon, Polyline, MultiPoint, "
                          "or Spatially enabled DataFrame"))
    return None


import arcgis
from arcgis._impl.common._utils import _lazy_property
from arcgis.apps.tracker import LocationTrackingError


class LocationTrackingManager:
    """
    This manages Location Tracking for an organization.
    It can be accessed from the gis as :py:attr:`~arcgis.gis.admin.location_tracking`

    Additional information can be found here: https://doc.arcgis.com/en/tracker/help/configure-location-tracking.htm

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    gis                    Required :class:`~arcgis.gis.GIS`. The GIS to configure location
                           tracking for.
    ==================     ====================================================================
    """
    def __init__(self, gis):
        self._gis = gis

    def _validate_environment(self, location_tracking_enabled=True):
        if self._gis is None:
            raise LocationTrackingError("You must use a valid GIS")
        if float(self._gis.properties.get("currentVersion", "0")) < 7.1:
            raise LocationTrackingError("Location Tracking requires ArcGIS Enterprise 10.7+ or ArcGIS Online")
        if self._gis.users.me.role != "org_admin":
            raise LocationTrackingError("You must be an administrator to configure Location Tracking")
        if self._gis.properties.get("isPortal", False) and not self._gis.properties.get("hasSpatioTemporalArcGISDataStore", False):
            raise LocationTrackingError("Location Tracking requires the Spatiotemporal Big Data Store")
        if location_tracking_enabled and "locationTracking" not in self._gis.properties.helperServices:
            raise LocationTrackingError("Location Tracking is not enabled.")
        if not location_tracking_enabled and "locationTracking" in self._gis.properties.helperServices:
            raise LocationTrackingError("Location Tracking is already enabled.")

    def enable(self):
        """
        Enables location tracking for the organization.

        :return: True if successful, False otherwise
        """
        if self.status == "enabled":
            return False
        self._validate_environment(location_tracking_enabled=False)
        for folder in self._gis.users.me.folders:
            if folder["title"] == "Location Tracking":
                break
        else:
            self._gis.content.create_folder("Location Tracking")
        item = self._gis.content.create_service(
            "location_tracking",
            create_params={
                "name": "location_tracking",
                "layers": [
                    {
                        "adminLayerInfo": {
                            "tableMetadata": {
                                "numberOfShards": "6",
                                "rollingIndexStrategy": "Yearly",
                                "dataRetentionStrategy": "30",
                                "dataRetentionStrategyUnits": "DAYS",
                                "dataRetention": "true"
                            }
                        }
                    },
                    {
                        "adminLayerInfo": {
                            "tableMetadata": {
                                "numberOfShards": "2",
                                "rollingIndexStrategy": "Yearly",
                                "dataRetentionStrategy": "30",
                                "dataRetentionStrategyUnits": "DAYS",
                                "dataRetention": "false"
                            }
                        }
                    }
                ],
                "description": "Location Tracking Service",
                "snippet": "Location Tracking Service"
            },
            folder="Location Tracking",
            service_type="locationTrackingService"
        )
        item.protect(True)
        item.share(org=True)
        self._gis.update_properties({
            "locationTrackingService": {
                "url": item.url,
                "id": item.itemid
            }
        })
        return True

    def pause(self):
        """
        Pauses location tracking for the organization.

        :return: True if successful, False otherwise
        """
        self._validate_environment()
        if self.status == "enabled":
            return bool(self._service.manager.update_definition({"capabilities": "Query"})["success"])
        else:
            return False

    def resume(self):
        """
        Resumes location tracking for the organization after it was paused.

        :return: True if successful, False otherwise
        """
        self._validate_environment()
        if self.status == "paused":
            return bool(self._service.manager.update_definition({"capabilities": "Query,Create,Update"})["success"])
        else:
            return False

    def disable(self):
        """
        Disables location tracking for the organization.

        THIS WILL DELETE ALL LOCATION TRACKING VIEWS, SERVICES, AND DATA.

        :return: True if successful, False otherwise
        """
        if self.status == "disabled":
            return False
        self._validate_environment()
        folder_id = self.item.ownerFolder
        views = self.item.related_items("Service2Service", "forward")
        for view in views:
            if hasattr(view, "properties") and view.properties is not None and "trackViewGroup" in view.properties:
                group = self._gis.groups.get(view.properties["trackViewGroup"])
                if group:
                    group.protected = False
                    group.delete()
            view.protect(False)
            view.delete()
        self.item.protect(False)
        self.item.delete()
        self._gis.update_properties({
            "locationTrackingService": "null"
        })
        if len(self._gis.content.search("""owner:"{}" ownerfolder:{}""".format(self.item.owner, folder_id))) == 0:
            for folder in self._gis.users.get(self.item.owner).folders:
                if folder["title"] == "Location Tracking":
                    self._gis.content.delete_folder(folder["title"], owner=self.item.owner)
                    break
        return True

    def create_track_view(self, title):
        """
        This creates a :class:`~arcgis.apps.tracker.TrackView`.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        title                  Required :class:`String` The title of the Track View to create.
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.tracker.TrackView`
        """
        self._validate_environment()
        folder = None
        if self.item.ownerFolder is not None:
            for f in self._gis.users.get(self.item.owner).folders:
                if f["id"] == self.item.ownerFolder:
                    folder = f["title"]
                    break
        group = self._gis.groups.create(title,
                                  "Location Tracking Group",
                                        is_view_only=True,
                                        is_invitation_only=True,
                                        access="private")
        if group.owner != self.item.owner:
            group.reassign_to(self.item.owner)
            group.remove_users([self._gis.users.me])
        group.protected = True
        item = self._gis.content.create_service(
            "{}_Track_View".format(group.id),
            create_params={
                "name": "{}_Track_View".format(group.id),
                "isView": True
            },
            folder=folder,
            service_type="locationTrackingService",
            is_view=True,
            owner=self.item.owner
        )
        item.protect(True)
        item.update({
            "title": title,
            "properties": {
                "trackViewGroup": group.id
            }
        })
        definition = {"viewDefinitionQuery": "created_user in ('')"}
        # Workaround for 10.7 bug where time wasn't enabled
        if self._gis.properties.isPortal and self._gis.version <= [7,1]:
            definition["timeInfo"] = {"startTimeField": "location_timestamp"}
        item.layers[0].manager.update_definition(definition)
        item.layers[1].manager.update_definition(definition)
        if self._gis.properties.isPortal:
            # Set allowOthersToQuery to True - workaround for missing feature in Enterprise 10.7/10.7.1
            arcgis.features.FeatureLayerCollection(item.url, self._gis).manager.update_definition(
                {
                    "editorTrackingInfo": {
                        "enableOwnershipAccessControl": True,
                        "enableEditorTracking": True,
                        "allowOthersToQuery": True,
                        "allowOthersToUpdate": False,
                        "allowOthersToDelete": False
                    }
                }
            )
        item.share(groups=[group])
        return arcgis.apps.tracker.TrackView(item)

    @property
    def retention_period(self):
        """
        The retention period of the Location Tracking Tracks Layer.
        This is a positive integer whose units are defined by :py:attr:`~arcgis.gis.admin.location_tracking.retention_period_units`
        """
        try:
            return int(self.tracks_layer.manager.properties["adminLayerInfo"]["tableMetadata"]["dataRetentionStrategy"])
        except:
            return None

    @retention_period.setter
    def retention_period(self, value):
        self._validate_environment()
        if isinstance(value, str):
            if not value.isdigit():
                raise LocationTrackingError("Invalid Retention Policy Setting: '{}' expected an integer greater than 0")
        elif isinstance(value, int):
            if value < 0:
                raise LocationTrackingError("Invalid Retention Policy Setting: '{}' expected an integer greater than 0")
        else:
            raise LocationTrackingError("Invalid Retention Policy Setting: '{}' expected an integer greater than 0")
        self.tracks_layer.manager.update_definition({
            "tableMetadata": {
                "dataRetentionStrategy": "{}".format(value)
            }
        })

    @property
    def retention_period_units(self):
        """The retention period units ("HOURS", "DAYS", "MONTHS", "YEARS") of the Location Tracking Tracks Layer"""
        try:
            return self.tracks_layer.manager.properties["adminLayerInfo"]["tableMetadata"]["dataRetentionStrategyUnits"]
        except:
            return None

    @retention_period_units.setter
    def retention_period_units(self, value):
        self._validate_environment()
        if not isinstance(value, str):
            raise LocationTrackingError("Invalid Retention Policy Setting: '{}' expected HOURS, DAYS, MONTHS, or YEARS".format(value))
        elif value.upper() not in ("HOURS", "DAYS", "MONTHS", "YEARS"):
            raise LocationTrackingError("Invalid Retention Policy Setting: '{}' expected HOURS, DAYS, MONTHS, or YEARS".format(value))
        self.tracks_layer.manager.update_definition({
            "tableMetadata": {
                "dataRetentionStrategyUnits": "{}".format(value)
            }
        })

    @property
    def retention_period_enabled(self):
        """A boolean indicating if the retention period is enabled"""
        try:
            return self.tracks_layer.manager.properties["adminLayerInfo"]["tableMetadata"]["dataRetention"].lower() == "true"
        except:
            return None

    @retention_period_enabled.setter
    def retention_period_enabled(self, value):
        self._validate_environment()
        if isinstance(value, str):
            if value.lower() not in ("false", "true"):
                raise LocationTrackingError("Invalid Retention Policy Setting: '{}' expected True or False".format(value))
        elif not isinstance(value, bool):
            raise LocationTrackingError("Invalid Retention Policy Setting: '{}' expected True or False".format(value))
        self.tracks_layer.manager.update_definition({
            "tableMetadata": {
                "dataRetention": "{}".format(str(value).lower())
            }
        })

    @_lazy_property
    def item(self):
        """The Location Tracking :class:`~arcgis.gis.Item`"""
        try:
            return self._gis.content.get(self._gis.properties.helperServices["locationTracking"]["id"])
        except:
            return None

    @_lazy_property
    def tracks_layer(self):
        """The tracks :class:`~arcgis.features.FeatureLayer`"""
        try:
            return self.item.layers[0]
        except:
            return None

    @_lazy_property
    def last_known_locations_layer(self):
        """The last known locations :class:`~arcgis.features.FeatureLayer`"""
        try:
            return self.item.layers[1]
        except:
            return None

    @property
    def status(self):
        """The status of location tracking ("disabled", "paused", "enabled")"""
        try:
            if "locationTracking" not in self._gis.properties.helperServices:
                return "disabled"
            if "Create" in self._service.properties["capabilities"]:
                return "enabled"
            else:
                return "paused"
        except:
            return "disabled"

    @_lazy_property
    def _service(self):
        try:
            return arcgis.features.FeatureLayerCollection(self._gis.properties.helperServices["locationTracking"]["url"], self._gis)
        except:
            return None



"""
This is the ArcGIS Notebook Server API Framework
"""

import os
from urllib.parse import urlparse

from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap

from ._logs import LogManager
from ._system import SystemManager
from ._security import SecurityManager
from ._machines import MachineManager
from ._nbm import NotebookManager

########################################################################
class NotebookServer(object):
    """
    Provides access to the ArcGIS Notebook Server administration API.
    """
    _gis = None
    _url = None
    _properties = None
    _logs = None
    _system = None
    _machine = None
    _notebook = None
    _security = None
    _version = None
    _sitemanager = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        if url.lower().endswith("/admin") == False:
            url += "/admin"
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
        else:
            raise ValueError("Invalid GIS object")
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {'f': 'json'}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return "<NotebookServer @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return "<NotebookServer @ {url}>".format(url=self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """Properties of the object"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    @property
    def version(self):
        """
        Returns the notebook server version

        :returns: List
        """
        if self._version is None:
            self._version = [int(i) for i in self.properties.version.split('.')]
        return self._version
    #----------------------------------------------------------------------
    @property
    def site(self):
        """
        Provides access to the notebook server's site management operations

        :returns: SiteManager
        """
        if self._sitemanager is None:
            from ._site import SiteManager
            self._sitemanager = SiteManager(url=self._url, notebook=self, gis=self._gis)
        return self._sitemanager
    #----------------------------------------------------------------------
    @property
    def info(self):
        """
        Returns information about the server site itself

        :returns: PropertyMap

        """
        url = self._url + "/info"
        params = {'f' : 'json'}
        res = self._gis._con.get(url, params)
        return PropertyMap(res)
    #----------------------------------------------------------------------
    @property
    def health_check(self):
        """

        The `health_check` verifies that your ArcGIS Notebook Server site
        has been created, and that its Docker environment has been
        correctly configured.

        **This is only avaible if the site can be accessed around the web adapter**

        :returns: boolean

        """
        netloc = urlparse(self._url).netloc
        url = "https://{base}:11443/arcgis/rest/info/healthcheck".format(base=netloc)
        params = {'f' : 'json'}
        res = self._gis._con.get(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    @property
    def logs(self):
        """
        Provides access to the notebook server's logging system

        :returns: LogManager

        """
        if self._logs is None:
            url = self._url + "/logs"
            self._logs = LogManager(url=url, gis=self._gis)
        return self._logs
    #----------------------------------------------------------------------
    @property
    def system(self):
        """
        returns access to the system properties of the ArcGIS Notebook Server

        :return: SystemManager

        """
        if self._system is None:
            url = self._url + "/system"
            self._system = SystemManager(url=url, gis=self._gis)
        return self._system
    #----------------------------------------------------------------------
    @property
    def machine(self):
        """
        Provides access to managing the registered machines with ArcGIS
        Notebook Server

        :returns: MachineManager

        """
        if self._machine is None:
            url = self._url + "/machines"
            self._machine = MachineManager(url=url, gis=self._gis)
        return self._machine
    #----------------------------------------------------------------------
    @property
    def security(self):
        """
        Provides access to managing the ArcGIS Notebook Server's security
        settings.

        :return: SecurityManager

        """
        if self._security is None:
            url = self._url + "/security"
            self._security = SecurityManager(url=url, gis=self._gis)
        return self._security
    #----------------------------------------------------------------------
    @property
    def notebooks(self):
        """
        Provides access to managing the ArcGIS Notebook Server's
        Notebooks

        :return: NotebookManager
        """
        if self._notebook is None:
            url = self._url + "/notebooks"
            self._notebook = NotebookManager(url=url, gis=self._gis)
        return self._notebook
    #----------------------------------------------------------------------
    @property
    def url(self):
        """The URL of the notebook server."""
        return self._url
